from datetime import datetime, timezone
from ipaddress import IPv4Address
from typing import Any, Dict, List, Optional
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import Mock, _Call, call, sentinel

from fred_types import ContactId, ContactRef, DomainId, KeysetId, KeysetRef, NssetId, NssetRef
from freezegun import freeze_time
from regal import (AdditionalIdentifier, Address, Contact, ContactClient, ContactIdentifierType, DnsHost, DnsKey,
                   Domain, DomainClient, Keyset, KeysetClient, Nsset, NssetClient, ObjectEvent, ObjectEvents, Registrar,
                   RegistrarClient)
from typist import SecretaryClient

from statementor.constants import Template
from statementor.main import Statementor, SyncStatementor


@freeze_time("2020-02-01 14:00:00")
class StatementorTest(IsolatedAsyncioTestCase):
    registrar = Registrar(registrar_handle='HOLLY', name='Holly', organization='JMC')
    events = ObjectEvents(
        registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
        transferred=ObjectEvent(registrar_handle='HOLLY'))
    pdf_data = b"Gazpacho!"

    def setUp(self):
        self.secretary_mock = Mock(spec=SecretaryClient)
        self.secretary_mock.render_pdf.return_value = self.pdf_data
        self.contact_mock = Mock(spec=ContactClient)
        self.domain_mock = Mock(spec=DomainClient)
        self.keyset_mock = Mock(spec=KeysetClient)
        self.nsset_mock = Mock(spec=NssetClient)
        self.registrar_mock = Mock(spec=RegistrarClient)
        self.registrar_mock.get_registrar_info.return_value = self.registrar

        self.statementor = Statementor(
            secretary_client=self.secretary_mock, contact_client=self.contact_mock, domain_client=self.domain_mock,
            keyset_client=self.keyset_mock, nsset_client=self.nsset_mock, registrar_client=self.registrar_mock)

    def _get_context(self, **kwargs: Any) -> Dict[str, Any]:
        registrar_context = {"handle": 'HOLLY', "name": 'Holly', "organization": 'JMC'}
        return dict({
            "private": sentinel.private,
            "timestamp": "2020-02-01T14:00:00+00:00",
            "registrar": registrar_context,
            "flags": [],
        }, **kwargs)

    def _get_contact_context(self, **kwargs: Any) -> Dict[str, Any]:
        context = self._get_context(name='', organization='', vat_identification_number=None, emails=[],
                                    notify_emails=[], telephone=None, fax=None, publish={},
                                    registered='1988-08-06T00:00:00+00:00')
        return dict(context, **kwargs)

    async def _test_get_contact_statement(self, contact: Contact, state: Dict[str, bool], context: Dict[str, Any],
                                          ) -> None:
        contact_id = ContactId('42')
        self.contact_mock.get_contact_info.return_value = contact
        self.contact_mock.get_contact_state.return_value = state

        result = await self.statementor.get_contact_statement(contact_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.contact_mock.mock_calls,
                         [call.get_contact_info(contact_id), call.get_contact_state(contact_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.CONTACT, context)])

    async def test_get_contact_statement(self):
        handle = "KRYTEN"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        context = self._get_contact_context(handle=handle)

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_flags(self):
        handle = "Starbug-1"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        context = self._get_contact_context(handle=handle, flags=['linked'])

        await self._test_get_contact_statement(contact, {'linked': True, 'unlinked': False}, context)

    async def test_get_contact_statement_data(self):
        # Test contact with most of its data
        handle = "KRYTEN"
        contact = Contact(
            contact_id=ContactId('42'), contact_handle=handle, name='Kryten', organization='JMC',
            vat_identification_number='2X4B-523P', emails=['kryten@example.org', 'abel@example.org'],
            notify_emails=['notify@example.org', 'holly@example.org'], telephone='+1.234', fax='+1.234',
            sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_contact_context(
            handle=handle, name='Kryten', organization='JMC', vat_identification_number='2X4B-523P',
            emails=['kryten@example.org', 'abel@example.org'],
            notify_emails=['notify@example.org', 'holly@example.org'], telephone='+1.234', fax='+1.234')

        await self._test_get_contact_statement(contact, {}, context)

    async def _test_get_contact_statement_ident(self, ident: ContactIdentifierType, kwargs: Dict[str, Any]) -> None:
        # Test additional identifiers.
        handle = "KRYTEN"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle,
                          additional_identifier=AdditionalIdentifier(type=ident, value='2X4B-523P'),
                          sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_contact_context(handle=handle, **kwargs)

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_national_ident_number(self):
        await self._test_get_contact_statement_ident(ContactIdentifierType.NATIONAL_IDENTITY_NUMBER,
                                                     {'national_identity_number': '2X4B-523P'})

    async def test_get_contact_statement_national_ident_card(self):
        await self._test_get_contact_statement_ident(ContactIdentifierType.NATIONAL_IDENTITY_CARD,
                                                     {'national_identity_card': '2X4B-523P'})

    async def test_get_contact_statement_passport_number(self):
        await self._test_get_contact_statement_ident(ContactIdentifierType.PASSPORT_NUMBER,
                                                     {'passport_number': '2X4B-523P'})

    async def test_get_contact_statement_company_reg_number(self):
        await self._test_get_contact_statement_ident(ContactIdentifierType.COMPANY_REGISTRATION_NUMBER,
                                                     {'company_registration_number': '2X4B-523P'})

    async def test_get_contact_statement_social_security_number(self):
        await self._test_get_contact_statement_ident(ContactIdentifierType.SOCIAL_SECURITY_NUMBER,
                                                     {'social_security_number': '2X4B-523P'})

    async def test_get_contact_statement_birthdate(self):
        # Birth date is passed as a string.
        await self._test_get_contact_statement_ident(ContactIdentifierType.BIRTHDATE,
                                                     {'birthdate': '2X4B-523P'})

    async def test_get_contact_statement_address(self):
        handle = "KRYTEN"
        address = Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC', country_code='JU',
                          state_or_province='Deep Space')
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, place=address,
                          sponsoring_registrar='HOLLY', events=self.events)
        address_context = {'streets': ['Deck 16'], 'city': 'Red Dwarf', 'postal_code': 'JMC', 'country_code': 'JU',
                           'state_or_province': 'Deep Space'}
        context = self._get_contact_context(handle=handle, address=address_context)

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_publish(self):
        handle = "KRYTEN"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          publish={'email': True, 'telephone': False}, events=self.events)
        context = self._get_contact_context(handle=handle, publish={'email': True, 'telephone': False})

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_publish_address(self):
        handle = "KRYTEN"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          publish={'place': True}, events=self.events)
        context = self._get_contact_context(handle=handle, publish={'address': True})

        await self._test_get_contact_statement(contact, {}, context)

    async def _test_get_contact_statement_publish_ident(self, ident: ContactIdentifierType, publish: bool,
                                                        publish_context: Dict[str, bool]) -> None:
        # Test publish of additional identifiers.
        handle = "KRYTEN"
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          additional_identifier=AdditionalIdentifier(type=ident, value='2X4B-523P'),
                          publish={'additional_identifier': publish}, events=self.events)
        context = self._get_contact_context(handle=handle, publish=publish_context, **{ident.value: '2X4B-523P'})

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_publish_ident_number(self):
        await self._test_get_contact_statement_publish_ident(ContactIdentifierType.NATIONAL_IDENTITY_NUMBER, True,
                                                             {'national_identity_number': True})

    async def test_get_contact_statement_publish_ident_number_false(self):
        await self._test_get_contact_statement_publish_ident(ContactIdentifierType.NATIONAL_IDENTITY_NUMBER, False,
                                                             {'national_identity_number': False})

    async def test_get_contact_statement_updated(self):
        handle = "KRYTEN"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY'),
            updated=ObjectEvent(registrar_handle='QUEEG-500', timestamp=datetime(1989, 11, 14, tzinfo=timezone.utc)))
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=events)
        context = self._get_contact_context(handle=handle, updated='1989-11-14T00:00:00+00:00')

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_updated_no_date(self):
        handle = "KRYTEN"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY'),
            updated=ObjectEvent(registrar_handle='QUEEG-500'))
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=events)
        context = self._get_contact_context(handle=handle)

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_transferred(self):
        handle = "KRYTEN"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY', timestamp=datetime(1988, 9, 6, tzinfo=timezone.utc)))
        contact = Contact(contact_id=ContactId('42'), contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=events)
        context = self._get_contact_context(handle=handle, transferred='1988-09-06T00:00:00+00:00')

        await self._test_get_contact_statement(contact, {}, context)

    async def test_get_contact_statement_handle(self):
        handle = "KRYTEN"
        contact_id = ContactId('42')
        contact = Contact(contact_id=contact_id, contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        self.contact_mock.get_contact_id.return_value = contact_id
        self.contact_mock.get_contact_info.return_value = contact
        self.contact_mock.get_contact_state.return_value = {}

        result = await self.statementor.get_contact_statement(handle, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        calls = [call.get_contact_id(handle), call.get_contact_info(contact_id), call.get_contact_state(contact_id)]
        self.assertEqual(self.contact_mock.mock_calls, calls)
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.CONTACT, self._get_contact_context(handle=handle))])

    async def test_get_contact_statement_timeout(self):
        handle = "KRYTEN"
        contact_id = ContactId('42')
        contact = Contact(contact_id=contact_id, contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        self.contact_mock.get_contact_info.return_value = contact
        self.contact_mock.get_contact_state.return_value = {}

        result = await self.statementor.get_contact_statement(contact_id, private=sentinel.private,
                                                              timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        context = self._get_contact_context(handle=handle)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.CONTACT, context, timeout=sentinel.timeout)])

    async def _test_get_domain_statement(
            self, domain: Domain, state: Dict[str, bool], context: Dict[str, Any], *,
            contacts: Optional[List[Contact]] = None, registrar_calls: Optional[List[_Call]] = None,
            contact_calls: Optional[List[_Call]] = None) -> None:
        domain_id = DomainId('42')
        self.domain_mock.get_domain_info.return_value = domain
        self.domain_mock.get_domain_state.return_value = state
        contacts = contacts or [Contact(contact_id=ContactId('42'), contact_handle='KRYTEN',
                                        sponsoring_registrar='HOLLY', events=self.events)]
        self.contact_mock.get_contact_info.side_effect = contacts
        contact_context: Dict[str, Any] = {
            'handle': 'KRYTEN', 'name': '', 'organization': '', 'vat_identification_number': None, 'emails': [],
            'notify_emails': [], 'telephone': None, 'fax': None, 'publish': {}}
        context.setdefault('registrant', contact_context)

        result = await self.statementor.get_domain_statement(domain_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.domain_mock.mock_calls,
                         [call.get_domain_info(domain_id), call.get_domain_state(domain_id)])
        self.assertEqual(self.contact_mock.mock_calls, contact_calls or [call.get_contact_info('2X4B')])
        self.assertEqual(self.registrar_mock.mock_calls, registrar_calls or [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.DOMAIN, context)])

    async def test_get_domain_statement(self):
        handle = "example.org"
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

    async def test_get_domain_statement_flags(self):
        handle = "example.org"
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_context(handle=handle, flags=['linked'], registered='1988-08-06T00:00:00+00:00',
                                    administrative_contacts=[])

        await self._test_get_domain_statement(domain, {'linked': True, 'unlinked': False}, context)

    async def test_get_domain_statement_admin(self):
        handle = "example.org"
        admin_id = ContactId('196')
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'), sponsoring_registrar='HOLLY',
                        administrative_contacts=[ContactRef(id=admin_id, handle='RIMMER')], events=self.events)
        admin_context = {"handle": 'RIMMER', "name": 'Arnold Rimmer', "organization": 'JMC administrative'}
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00',
                                    administrative_contacts=[admin_context])
        contacts = [
            Contact(contact_id=ContactId('42'), contact_handle='KRYTEN', sponsoring_registrar='HOLLY',
                    events=self.events),
            Contact(contact_id=admin_id, contact_handle='RIMMER', name='Arnold Rimmer',
                    organization='JMC administrative')]

        await self._test_get_domain_statement(
            domain, {}, context, contacts=contacts,
            contact_calls=[call.get_contact_info('2X4B'), call.get_contact_info(admin_id)])

    async def test_get_domain_statement_nsset(self):
        handle = "example.org"
        nsset_id = NssetId('196')
        nsset_handle = 'Starbug-1'
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        nsset=NssetRef(id=nsset_id, handle=nsset_handle), sponsoring_registrar='HOLLY',
                        events=self.events)
        self.nsset_mock.get_nsset_info.return_value = Nsset(nsset_id=NssetId('42'), nsset_handle=nsset_handle,
                                                            sponsoring_registrar='HOLLY')
        self.nsset_mock.get_nsset_state.return_value = {'nsset_flag': True}
        nsset_context = {'handle': 'Starbug-1', 'nameservers': [], 'technical_contacts': [],
                         'registrar': {'handle': 'HOLLY', 'name': 'Holly', 'organization': 'JMC'},
                         'flags': ['nsset_flag']}
        context = self._get_context(handle=handle, nsset=nsset_context, registered='1988-08-06T00:00:00+00:00',
                                    administrative_contacts=[])

        await self._test_get_domain_statement(
            domain, {}, context,
            registrar_calls=[call.get_registrar_info('HOLLY'), call.get_registrar_info('HOLLY')])

        self.assertEqual(self.nsset_mock.mock_calls, [call.get_nsset_info(nsset_id), call.get_nsset_state(nsset_id)])

    async def test_get_domain_statement_keyset(self):
        handle = "example.org"
        keyset_id = KeysetId('196')
        keyset_handle = 'Starbug-1'
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        keyset=KeysetRef(id=keyset_id, handle=keyset_handle), sponsoring_registrar='HOLLY',
                        events=self.events)
        self.keyset_mock.get_keyset_info.return_value = Keyset(keyset_id=KeysetId('42'), keyset_handle=keyset_handle,
                                                               sponsoring_registrar='HOLLY')
        self.keyset_mock.get_keyset_state.return_value = {'keyset_flag': True}
        context = self._get_context(handle=handle,
                                    keyset={'handle': 'Starbug-1', 'dns_keys': [], 'flags': ['keyset_flag']},
                                    registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

        self.assertEqual(self.keyset_mock.mock_calls,
                         [call.get_keyset_info(keyset_id), call.get_keyset_state(keyset_id)])

    async def test_get_domain_statement_expires_at(self):
        handle = "example.org"
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        expires_at=datetime(2020, 4, 9, tzinfo=timezone.utc), sponsoring_registrar='HOLLY',
                        events=self.events)
        context = self._get_context(handle=handle, expires_at='2020-04-09T00:00:00+00:00',
                                    registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

    async def test_get_domain_statement_updated(self):
        handle = "example.org"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY'),
            updated=ObjectEvent(registrar_handle='QUEEG-500', timestamp=datetime(1989, 11, 14, tzinfo=timezone.utc)))
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=events)
        context = self._get_context(handle=handle, updated='1989-11-14T00:00:00+00:00',
                                    registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

    async def test_get_domain_statement_updated_no_date(self):
        handle = "example.org"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY'),
            updated=ObjectEvent(registrar_handle='QUEEG-500'))
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=events)
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

    async def test_get_domain_statement_transferred(self):
        handle = "example.org"
        events = ObjectEvents(
            registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
            transferred=ObjectEvent(registrar_handle='HOLLY', timestamp=datetime(1988, 9, 6, tzinfo=timezone.utc)))
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=events)
        context = self._get_context(handle=handle, transferred='1988-09-06T00:00:00+00:00',
                                    registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        await self._test_get_domain_statement(domain, {}, context)

    async def test_get_domain_statement_handle(self):
        handle = "example.org"
        domain_id = DomainId('42')
        domain = Domain(domain_id=domain_id, fqdn=handle, registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        self.domain_mock.get_domain_id.return_value = domain_id
        self.domain_mock.get_domain_info.return_value = domain
        self.contact_mock.get_contact_info.return_value = Contact(contact_id=ContactId('42'), contact_handle='KRYTEN',
                                                                  sponsoring_registrar='HOLLY', events=self.events)
        self.domain_mock.get_domain_state.return_value = {}

        result = await self.statementor.get_domain_statement(handle, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        calls = [call.get_domain_id(handle), call.get_domain_info(domain_id), call.get_domain_state(domain_id)]
        self.assertEqual(self.domain_mock.mock_calls, calls)
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        contact_context: Dict[str, Any] = {
            'handle': 'KRYTEN', 'name': '', 'organization': '', 'vat_identification_number': None, 'emails': [],
            'notify_emails': [], 'telephone': None, 'fax': None, 'publish': {}}
        context = self._get_context(handle=handle, registrant=contact_context, registered='1988-08-06T00:00:00+00:00',
                                    administrative_contacts=[])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.DOMAIN, context)])

    async def test_get_domain_statement_timeout(self):
        handle = "KRYTEN"
        domain_id = DomainId('42')
        domain = Domain(domain_id=DomainId('42'), fqdn=handle,
                        registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        self.domain_mock.get_domain_info.return_value = domain
        self.domain_mock.get_domain_state.return_value = {}
        self.contact_mock.get_contact_info.return_value = Contact(contact_id=ContactId('42'), contact_handle='KRYTEN',
                                                                  sponsoring_registrar='HOLLY', events=self.events)

        result = await self.statementor.get_domain_statement(domain_id, private=sentinel.private,
                                                             timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        contact_context: Dict[str, Any] = {
            'handle': 'KRYTEN', 'name': '', 'organization': '', 'vat_identification_number': None, 'emails': [],
            'notify_emails': [], 'telephone': None, 'fax': None, 'publish': {}}
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])
        context.setdefault('registrant', contact_context)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.DOMAIN, context, timeout=sentinel.timeout)])

    async def _test_get_keyset_statement(self, keyset: Keyset, state: Dict[str, bool], context: Dict[str, Any]) -> None:
        keyset_id = KeysetId('42')
        self.keyset_mock.get_keyset_info.return_value = keyset
        self.keyset_mock.get_keyset_state.return_value = state

        result = await self.statementor.get_keyset_statement(keyset_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.keyset_mock.mock_calls,
                         [call.get_keyset_info(keyset_id), call.get_keyset_state(keyset_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.KEYSET, context)])

    async def test_get_keyset_statement(self):
        handle = "Starbug-1"
        keyset = Keyset(keyset_id=KeysetId('42'), keyset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[])

        await self._test_get_keyset_statement(keyset, {}, context)

    async def test_get_keyset_statement_flags(self):
        handle = "Starbug-1"
        keyset = Keyset(keyset_id=KeysetId('42'), keyset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[], flags=['linked'])

        await self._test_get_keyset_statement(keyset, {'linked': True, 'unlinked': False}, context)

    async def test_get_keyset_statement_dns_key(self):
        handle = "Starbug-1"
        keyset = Keyset(keyset_id=KeysetId('42'), keyset_handle=handle,
                        dns_keys=[DnsKey(flags=128, protocol=3, alg=7, key="GAZPACHO!")],
                        sponsoring_registrar='HOLLY')
        dns_key = {"flags": 128, "protocol": 3, "algorithm": 7, "key": "GAZPACHO!"}
        context = self._get_context(handle=handle, dns_keys=[dns_key], technical_contacts=[])

        await self._test_get_keyset_statement(keyset, {}, context)

    async def test_get_keyset_statement_tech(self):
        handle = "Starbug-1"
        tech_id = ContactId('196')
        keyset = Keyset(keyset_id=KeysetId('42'), keyset_handle=handle,
                        technical_contacts=[ContactRef(id=tech_id, handle='RIMMER')], sponsoring_registrar='HOLLY')
        tech_context = {"handle": 'RIMMER', "name": 'Arnold Rimmer', "organization": 'JMC technical'}
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[tech_context])
        self.contact_mock.get_contact_info.return_value = Contact(
            contact_id=tech_id, contact_handle='RIMMER', name='Arnold Rimmer', organization='JMC technical')

        await self._test_get_keyset_statement(keyset, {}, context)

        self.assertEqual(self.contact_mock.mock_calls, [call.get_contact_info(tech_id)])

    async def test_get_keyset_statement_handle(self):
        handle = "example.org"
        keyset_id = KeysetId('42')
        keyset = Keyset(keyset_id=keyset_id, keyset_handle=handle, sponsoring_registrar='HOLLY')
        self.keyset_mock.get_keyset_id.return_value = keyset_id
        self.keyset_mock.get_keyset_info.return_value = keyset
        self.keyset_mock.get_keyset_state.return_value = {}

        result = await self.statementor.get_keyset_statement(handle, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        calls = [call.get_keyset_id(handle), call.get_keyset_info(keyset_id), call.get_keyset_state(keyset_id)]
        self.assertEqual(self.keyset_mock.mock_calls, calls)
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.KEYSET, context)])

    async def test_get_keyset_statement_timeout(self):
        handle = "Starbug-1"
        keyset_id = KeysetId('42')
        keyset = Keyset(keyset_id=keyset_id, keyset_handle=handle, sponsoring_registrar='HOLLY')
        self.keyset_mock.get_keyset_info.return_value = keyset
        self.keyset_mock.get_keyset_state.return_value = {}

        result = await self.statementor.get_keyset_statement(keyset_id, private=sentinel.private,
                                                             timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[])
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.KEYSET, context, timeout=sentinel.timeout)])

    async def _test_get_nsset_statement(self, nsset: Nsset, state: Dict[str, bool], context: Dict[str, Any]) -> None:
        nsset_id = NssetId('42')
        self.nsset_mock.get_nsset_info.return_value = nsset
        self.nsset_mock.get_nsset_state.return_value = state

        result = await self.statementor.get_nsset_statement(nsset_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.nsset_mock.mock_calls, [call.get_nsset_info(nsset_id), call.get_nsset_state(nsset_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.NSSET, context)])

    async def test_get_nsset_statement(self):
        handle = "Starbug-1"
        nsset = Nsset(nsset_id=NssetId('42'), nsset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[])

        await self._test_get_nsset_statement(nsset, {}, context)

    async def test_get_nsset_statement_flags(self):
        handle = "Starbug-1"
        nsset = Nsset(nsset_id=NssetId('42'), nsset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[], flags=['linked'])

        await self._test_get_nsset_statement(nsset, {'linked': True, 'unlinked': False}, context)

    async def test_get_nsset_statement_nameserver(self):
        handle = "Starbug-1"
        nsset = Nsset(nsset_id=NssetId('42'), nsset_handle=handle,
                      dns_hosts=[DnsHost(fqdn='example.org', ip_addresses=[])],
                      sponsoring_registrar='HOLLY')
        nameserver = {"fqdn": 'example.org', "ip_addresses": []}
        context = self._get_context(handle=handle, nameservers=[nameserver], technical_contacts=[])

        await self._test_get_nsset_statement(nsset, {}, context)

    async def test_get_nsset_statement_nameserver_ip_address(self):
        handle = "Starbug-1"
        nsset = Nsset(nsset_id=NssetId('42'), nsset_handle=handle,
                      dns_hosts=[DnsHost(fqdn='example.org', ip_addresses=[IPv4Address('127.0.0.1')])],
                      sponsoring_registrar='HOLLY')
        nameserver = {"fqdn": 'example.org', "ip_addresses": ['127.0.0.1']}
        context = self._get_context(handle=handle, nameservers=[nameserver], technical_contacts=[])

        await self._test_get_nsset_statement(nsset, {}, context)

    async def test_get_nsset_statement_tech(self):
        handle = "Starbug-1"
        tech_id = ContactId('196')
        nsset = Nsset(nsset_id=NssetId('42'), nsset_handle=handle,
                      technical_contacts=[ContactRef(id=tech_id, handle='RIMMER')], sponsoring_registrar='HOLLY')
        tech_context = {"handle": 'RIMMER', "name": 'Arnold Rimmer', "organization": 'JMC technical'}
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[tech_context])
        self.contact_mock.get_contact_info.return_value = Contact(
            contact_id=tech_id, contact_handle='RIMMER', name='Arnold Rimmer', organization='JMC technical')

        await self._test_get_nsset_statement(nsset, {}, context)

        self.assertEqual(self.contact_mock.mock_calls, [call.get_contact_info(tech_id)])

    async def test_get_nsset_statement_handle(self):
        handle = "example.org"
        nsset_id = NssetId('42')
        nsset = Nsset(nsset_id=nsset_id, nsset_handle=handle, sponsoring_registrar='HOLLY')
        self.nsset_mock.get_nsset_id.return_value = nsset_id
        self.nsset_mock.get_nsset_info.return_value = nsset
        self.nsset_mock.get_nsset_state.return_value = {}

        result = await self.statementor.get_nsset_statement(handle, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        calls = [call.get_nsset_id(handle), call.get_nsset_info(nsset_id), call.get_nsset_state(nsset_id)]
        self.assertEqual(self.nsset_mock.mock_calls, calls)
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.NSSET, context)])

    async def test_get_nsset_statement_timeout(self):
        handle = "Starbug-1"
        nsset_id = NssetId('42')
        nsset = Nsset(nsset_id=nsset_id, nsset_handle=handle, sponsoring_registrar='HOLLY')
        self.nsset_mock.get_nsset_info.return_value = nsset
        self.nsset_mock.get_nsset_state.return_value = {}

        result = await self.statementor.get_nsset_statement(nsset_id, private=sentinel.private,
                                                            timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[])
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.NSSET, context, timeout=sentinel.timeout)])


@freeze_time("2020-02-01 14:00:00")
class SyncStatementorTest(TestCase):
    registrar = Registrar(registrar_handle='HOLLY', name='Holly', organization='JMC')
    events = ObjectEvents(
        registered=ObjectEvent(registrar_handle='DivaDroid', timestamp=datetime(1988, 8, 6, tzinfo=timezone.utc)),
        transferred=ObjectEvent(registrar_handle='HOLLY'))
    pdf_data = b"Gazpacho!"

    def setUp(self):
        self.secretary_mock = Mock(spec=SecretaryClient)
        self.secretary_mock.render_pdf.return_value = self.pdf_data
        self.contact_mock = Mock(spec=ContactClient)
        self.domain_mock = Mock(spec=DomainClient)
        self.keyset_mock = Mock(spec=KeysetClient)
        self.nsset_mock = Mock(spec=NssetClient)
        self.registrar_mock = Mock(spec=RegistrarClient)
        self.registrar_mock.get_registrar_info.return_value = self.registrar

        self.statementor = SyncStatementor(
            secretary_client=self.secretary_mock, contact_client=self.contact_mock, domain_client=self.domain_mock,
            keyset_client=self.keyset_mock, nsset_client=self.nsset_mock, registrar_client=self.registrar_mock)

    def _get_context(self, **kwargs: Any) -> Dict[str, Any]:
        registrar_context = {"handle": 'HOLLY', "name": 'Holly', "organization": 'JMC'}
        return dict({
            "private": sentinel.private,
            "timestamp": "2020-02-01T14:00:00+00:00",
            "registrar": registrar_context,
            "flags": [],
        }, **kwargs)

    def _get_contact_context(self, **kwargs: Any) -> Dict[str, Any]:
        context = self._get_context(name='', organization='', vat_identification_number=None, emails=[],
                                    notify_emails=[], telephone=None, fax=None, publish={},
                                    registered='1988-08-06T00:00:00+00:00')
        return dict(context, **kwargs)

    def test_get_contact_statement(self):
        handle = "KRYTEN"
        contact_id = ContactId('42')
        contact = Contact(contact_id=contact_id, contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        context = self._get_contact_context(handle=handle)

        self.contact_mock.get_contact_info.return_value = contact
        self.contact_mock.get_contact_state.return_value = {}

        result = self.statementor.get_contact_statement(contact_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.contact_mock.mock_calls,
                         [call.get_contact_info(contact_id), call.get_contact_state(contact_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.CONTACT, context)])

    def test_get_contact_statement_timeout(self):
        handle = "KRYTEN"
        contact_id = ContactId('42')
        contact = Contact(contact_id=contact_id, contact_handle=handle, sponsoring_registrar='HOLLY',
                          events=self.events)
        context = self._get_contact_context(handle=handle)

        self.contact_mock.get_contact_info.return_value = contact
        self.contact_mock.get_contact_state.return_value = {}

        result = self.statementor.get_contact_statement(contact_id, private=sentinel.private, timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.CONTACT, context, timeout=sentinel.timeout)])

    def test_get_domain_statement(self):
        handle = "example.org"
        domain_id = DomainId('42')
        domain = Domain(domain_id=domain_id, fqdn=handle, registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        self.domain_mock.get_domain_info.return_value = domain
        self.domain_mock.get_domain_state.return_value = {}
        self.contact_mock.get_contact_info.return_value = Contact(contact_id=ContactId('42'), contact_handle='KRYTEN',
                                                                  sponsoring_registrar='HOLLY')
        contact_context: Dict[str, Any] = {
            'handle': 'KRYTEN', 'name': '', 'organization': '', 'vat_identification_number': None, 'emails': [],
            'notify_emails': [], 'telephone': None, 'fax': None, 'publish': {}}
        context.setdefault('registrant', contact_context)

        result = self.statementor.get_domain_statement(domain_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.domain_mock.mock_calls,
                         [call.get_domain_info(domain_id), call.get_domain_state(domain_id)])
        self.assertEqual(self.contact_mock.mock_calls, [call.get_contact_info('2X4B')])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.DOMAIN, context)])

    def test_get_domain_statement_timeout(self):
        handle = "example.org"
        domain_id = DomainId('42')
        domain = Domain(domain_id=domain_id, fqdn=handle, registrant=ContactRef(id=ContactId('2X4B'), handle='KRYTEN'),
                        sponsoring_registrar='HOLLY', events=self.events)
        context = self._get_context(handle=handle, registered='1988-08-06T00:00:00+00:00', administrative_contacts=[])

        self.domain_mock.get_domain_info.return_value = domain
        self.domain_mock.get_domain_state.return_value = {}
        self.contact_mock.get_contact_info.return_value = Contact(contact_id=ContactId('42'), contact_handle='KRYTEN',
                                                                  sponsoring_registrar='HOLLY')
        contact_context: Dict[str, Any] = {
            'handle': 'KRYTEN', 'name': '', 'organization': '', 'vat_identification_number': None, 'emails': [],
            'notify_emails': [], 'telephone': None, 'fax': None, 'publish': {}}
        context.setdefault('registrant', contact_context)

        result = self.statementor.get_domain_statement(domain_id, private=sentinel.private, timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.DOMAIN, context, timeout=sentinel.timeout)])

    def test_get_keyset_statement(self):
        handle = "Starbug-1"
        keyset_id = KeysetId('42')
        keyset = Keyset(keyset_id=keyset_id, keyset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[])

        self.keyset_mock.get_keyset_info.return_value = keyset
        self.keyset_mock.get_keyset_state.return_value = {}

        result = self.statementor.get_keyset_statement(keyset_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.keyset_mock.mock_calls,
                         [call.get_keyset_info(keyset_id), call.get_keyset_state(keyset_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.KEYSET, context)])

    def test_get_keyset_statement_timeout(self):
        handle = "Starbug-1"
        keyset_id = KeysetId('42')
        keyset = Keyset(keyset_id=keyset_id, keyset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, dns_keys=[], technical_contacts=[])

        self.keyset_mock.get_keyset_info.return_value = keyset
        self.keyset_mock.get_keyset_state.return_value = {}

        result = self.statementor.get_keyset_statement(keyset_id, private=sentinel.private, timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.KEYSET, context, timeout=sentinel.timeout)])

    def test_get_nsset_statement(self):
        handle = "Starbug-1"
        nsset_id = NssetId('42')
        nsset = Nsset(nsset_id=nsset_id, nsset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[])

        self.nsset_mock.get_nsset_info.return_value = nsset
        self.nsset_mock.get_nsset_state.return_value = {}

        result = self.statementor.get_nsset_statement(nsset_id, private=sentinel.private)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.nsset_mock.mock_calls, [call.get_nsset_info(nsset_id), call.get_nsset_state(nsset_id)])
        self.assertEqual(self.registrar_mock.mock_calls, [call.get_registrar_info('HOLLY')])
        self.assertEqual(self.secretary_mock.mock_calls, [call.render_pdf(Template.NSSET, context)])

    def test_get_nsset_statement_timeout(self):
        handle = "Starbug-1"
        nsset_id = NssetId('42')
        nsset = Nsset(nsset_id=nsset_id, nsset_handle=handle, sponsoring_registrar='HOLLY')
        context = self._get_context(handle=handle, nameservers=[], technical_contacts=[])

        self.nsset_mock.get_nsset_info.return_value = nsset
        self.nsset_mock.get_nsset_state.return_value = {}

        result = self.statementor.get_nsset_statement(nsset_id, private=sentinel.private, timeout=sentinel.timeout)

        self.assertEqual(result.getvalue(), self.pdf_data)
        self.assertEqual(self.secretary_mock.mock_calls,
                         [call.render_pdf(Template.NSSET, context, timeout=sentinel.timeout)])
